
#include <w32api/windows.h>
#include <w32api/imagehlp.h>
#include <stdio.h>
#include <getopt.h>

int verbose = 0; 

int show_help(void)
{
	printf(
		"MSVC++ demangler by Dipl.-Ing. Ralf Habacker Germany <ralf.habacker@freenet.de>\n"
		"\nSyntax:\n"
		"\tvcdmgl [options] mangled_symbol_name \n"
		"\tvcdmgl [options] <infile\n"
		"\tcat infile | vcdmgl [options]\n"
		"\nOptions:\n"
  	"\t-v --verbose       print mangled symbol name too\n"
  	"\t-h --help\n\n"
		"\nExamples:\n"
		"\t$ ./vcdmgl ??0QMapPrivateBase@@QAE@PBV0@@Z <cr>\n"
		"\tpublic: __thiscall QMapPrivateBase::QMapPrivateBase(class QMapPrivateBase const *)\n\n"
		"\t$ nm /c/Programme/qt/lib/qt-mt230nc.lib  | gawk '$0 ~ /T \\\?/ { print $3 }' | ./vcdmgl  -v <cr>\n"
		"\t??0?$QArray@D@@IAE@HH@Z ->      protected: __thiscall QArray<char>::QArray<char>(int,int)\n"
		"\t??0?$QArray@D@@QAE@ABV0@@Z      ->      public: __thiscall QArray<char>::QArray<char>(class QArray<char> const &)\n"
		"\t...\n"
	);
}

char *short_options = "vh"; 

struct option long_options[] =
{
  {"verbose", no_argument, 0, 'v'},
  {"help", no_argument, 0, 'h'},
  {0, no_argument, 0, 0}
};

int msvc_demangle(char *s)
{
	char outbuf[1024];
	char *p,*q;
	int ret; 

	// eat trailing spaces 
	p = s;
	while(*p && isspace(*p))
		p++;
	// remove tail 
	if ((q = strchr(p,' ')) != 0)
		*q = '\0';
	// remove line feeds 
	else if ((q = strchr(p,'\n')) != 0)
		*q = '\0';
	ret = UnDecorateSymbolName(p,outbuf,1023,UNDNAME_COMPLETE | UNDNAME_32_BIT_DECODE);

	if (verbose) {
		if (ret > 0) 
			printf("%s\t->\t%s\n",p,outbuf);
		else 
			printf("demangling error on symbol '%s'\n",p);
	}
	else 
			printf("%s\n",outbuf);
	
	return ret > 0; 
}
int main(int argc,char **argv)
{
	char c; 
	FILE *f = NULL; 
  while ((c = getopt_long (argc, argv, short_options, long_options,(int *) 0)) != EOF)
    switch (c) {
			case 'v': 
				verbose = 1;
				break;
				
			case 'h': 
			default:
				show_help();
				exit(1);
		}
	if (optind < argc) {
		for (; optind < argc;) {
      char *name = argv[optind++];
			if (*name == '-')
				f = stdin; 
			else if (name) 
				msvc_demangle(name);
		}
	}
	else {
		f = stdin; 
	}
	if (f) {
		while (!feof(f)) {
			char inbuf[1024]; 
			fgets(inbuf,1023,f);
			msvc_demangle(inbuf);
		}
		if (f != stdin)
			fclose(f); 
	}

	return 0;
}
